from __future__ import annotations

import asyncio
import base64
import json
import os
import subprocess
from pathlib import Path
from pprint import pprint
from secrets import token_urlsafe

import argocd_client
import yaml
from kubernetes import client, config
from kubernetes.client.rest import ApiException
from omegaconf import OmegaConf
from pyhelm3 import Client

from MAIA.maia_fn import (
    convert_username_to_jupyterhub_username,
    encode_docker_registry_secret,
    generate_human_memorable_password,
    get_ssh_port_dict,
    get_ssh_ports,
)


def generate_minio_configs(namespace):
    """
    Generate configuration settings for MinIO.

    Parameters
    ----------
    namespace : int or str
        The unique identifier for the project.

    Returns
    -------
    dict
        A dictionary with the following keys:
        - access_key (str): The access key for MinIO.
        - secret_key (str): A randomly generated secret key for MinIO.
        - console_access_key (str): A base64 encoded access key for console access.
        - console_secret_key (str): A base64 encoded secret key for console access.
    """

    existing_minio_configs = get_minio_config_if_exists(namespace)
    minio_configs = {
        "access_key": "admin",
        "secret_key": (
            existing_minio_configs["secret_key"]
            if "secret_key" in existing_minio_configs
            else token_urlsafe(16).replace("-", "_")
        ),
        "console_access_key": (
            base64.b64encode(existing_minio_configs["console_access_key"].encode("ascii")).decode("ascii")
            if "console_access_key" in existing_minio_configs
            else base64.b64encode(token_urlsafe(16).replace("-", "_").encode("ascii")).decode("ascii")
        ),
        "console_secret_key": (
            base64.b64encode(existing_minio_configs["console_secret_key"].encode("ascii")).decode("ascii")
            if "console_secret_key" in existing_minio_configs
            else base64.b64encode(token_urlsafe(16).replace("-", "_").encode("ascii")).decode("ascii")
        ),
    }

    return minio_configs


def get_minio_config_if_exists(project_id):
    """
    Retrieves MinIO configuration if it exists for the given project ID.
    This function loads the Kubernetes configuration from the environment,
    accesses the Kubernetes API to list secrets in the specified namespace,
    and extracts MinIO-related configuration from the secrets.

    Parameters
    ----------
    project_id : str
        The ID of the project for which to retrieve the MinIO configuration.

    Returns
    -------
    dict
        A dictionary containing MinIO configuration keys and their corresponding values.
        The dictionary may contain the following keys:
        - "access_key": The default access key (always "admin").
        - "console_access_key": The console access key, if found.
        - "console_secret_key": The console secret key, if found.
        - "secret_key": The MinIO root password, if found.
    """
    if "KUBECONFIG_LOCAL" not in os.environ:
        os.environ["KUBECONFIG_LOCAL"] = os.environ["KUBECONFIG"]
    kubeconfig = yaml.safe_load(Path(os.environ["KUBECONFIG_LOCAL"]).read_text())
    config.load_kube_config_from_dict(kubeconfig)

    v1 = client.CoreV1Api()
    minio_configs = {"access_key": "admin"}
    secrets = v1.list_namespaced_secret(namespace=project_id.lower().replace("_", "-"))
    for secret in secrets.items:
        if secret.metadata.name == "storage-user":
            for item in secret.data:

                decoded_value = base64.b64decode(secret.data[item]).decode("ascii")
                if item == "CONSOLE_ACCESS_KEY":
                    minio_configs["console_access_key"] = decoded_value
                if item == "CONSOLE_SECRET_KEY":
                    minio_configs["console_secret_key"] = decoded_value
        if secret.metadata.name == "storage-configuration":
            for _, value in secret.data.items():
                decoded_value = base64.b64decode(value).decode("ascii")
                for line in decoded_value.split("\n"):
                    if line.startswith("export MINIO_ROOT_PASSWORD="):
                        minio_configs["secret_key"] = line[len("export MINIO_ROOT_PASSWORD=") :]

    return minio_configs


def generate_mlflow_configs(namespace):
    """
    Generate MLflow configuration dictionary with encoded user and password.

    Parameters
    ----------
    namespace : str
        The namespace to be encoded as the MLflow user.

    Returns
    -------
    dict
        A dictionary containing the encoded MLflow user and password.
    """
    existing_mlflow_configs = get_mlflow_config_if_exists(namespace)

    mlflow_configs = {
        "mlflow_user": (
            base64.b64encode(existing_mlflow_configs["mlflow_user"].encode("ascii")).decode("ascii")
            if "mlflow_user" in existing_mlflow_configs
            else base64.b64encode(namespace.encode("ascii")).decode("ascii")
        ),
        "mlflow_password": (
            base64.b64encode(existing_mlflow_configs["mlflow_password"].replace("-", "_").encode("ascii")).decode("ascii")
            if "mlflow_password" in existing_mlflow_configs
            else base64.b64encode(token_urlsafe(16).replace("-", "_").encode("ascii")).decode("ascii")
        ),
    }

    return mlflow_configs


def get_mlflow_config_if_exists(project_id):
    """
    Retrieve MLflow configuration from Kubernetes secrets if they exist.

    Parameters
    ----------
    project_id : str
        The ID of the project for which to retrieve the MLflow configuration. This ID is used to
        locate the corresponding Kubernetes namespace and secrets.

    Returns
    -------
    dict
        A dictionary containing the MLflow configuration with keys "mlflow_user" and "mlflow_password"
        if they exist in the Kubernetes secrets. If the secrets are not found, an empty dictionary is returned.

    Raises
    ------
    KeyError
        If the "KUBECONFIG" environment variable is not set.
    yaml.YAMLError
        If there is an error parsing the Kubernetes configuration file.
    kubernetes.client.exceptions.ApiException
        If there is an error communicating with the Kubernetes API.
    """
    if "KUBECONFIG_LOCAL" not in os.environ:
        os.environ["KUBECONFIG_LOCAL"] = os.environ["KUBECONFIG"]
    kubeconfig = yaml.safe_load(Path(os.environ["KUBECONFIG_LOCAL"]).read_text())
    config.load_kube_config_from_dict(kubeconfig)

    v1 = client.CoreV1Api()
    mlflow_configs = {}
    secrets = v1.list_namespaced_secret(namespace=project_id.lower().replace("_", "-"))
    for secret in secrets.items:

        if secret.metadata.name == project_id.lower().replace("_", "-"):
            for item in secret.data:

                decoded_value = base64.b64decode(secret.data[item]).decode("ascii")
                if item == "user":
                    mlflow_configs["mlflow_user"] = decoded_value
                if item == "password":
                    mlflow_configs["mlflow_password"] = decoded_value

    return mlflow_configs


def generate_mysql_configs(namespace):
    """
    Generate MySQL configuration dictionary.

    Parameters
    ----------
    namespace : str
        The namespace to be used as the MySQL user.

    Returns
    -------
    dict
        A dictionary containing MySQL user and password.
    """

    existing_mysql_configs = get_mysql_config_if_exists(namespace)

    mysql_configs = {
        "mysql_user": namespace,
        "mysql_password": (
            "".join(filter(str.isalnum, existing_mysql_configs["mysql_password"]))
            if "mysql_password" in existing_mysql_configs
            else "".join(filter(str.isalnum, token_urlsafe(16)))
        ),
    }

    return mysql_configs


def get_mysql_config_if_exists(project_id):
    """
    Retrieves MySQL configuration from Kubernetes environment variables if they exist.

    Parameters
    ----------
    project_id : str
        The ID of the project for which to retrieve the MySQL configuration. This ID is used to
        identify the namespace and the MySQL deployment within the Kubernetes cluster.

    Returns
    -------
    dict
        A dictionary containing the MySQL user and password if they exist in the environment
        variables of the MySQL deployment. The dictionary keys are:
        - "mysql_user": The MySQL user.
        - "mysql_password": The MySQL password.

    Notes
    -----
    This function assumes that the Kubernetes configuration file is specified in the environment
    variable "KUBECONFIG" and that the MySQL deployment name starts with the project ID followed
    by "-mysql-mkg".
    """
    if "KUBECONFIG_LOCAL" not in os.environ:
        os.environ["KUBECONFIG_LOCAL"] = os.environ["KUBECONFIG"]
    kubeconfig = yaml.safe_load(Path(os.environ["KUBECONFIG_LOCAL"]).read_text())
    config.load_kube_config_from_dict(kubeconfig)

    v1 = client.CoreV1Api()
    mlflow_configs = {}
    deploy = v1.list_namespaced_pod(namespace=project_id.lower().replace("_", "-"))

    for deployment in deploy.items:
        if deployment.metadata.name.startswith(project_id.lower().replace("_", "-") + "-mysql-mkg"):
            envs = deployment.spec.containers[0].env
            for env in envs:
                if env.name == "MYSQL_USER":
                    mlflow_configs["mysql_user"] = env.value
                if env.name == "MYSQL_PASSWORD":
                    mlflow_configs["mysql_password"] = env.value

    return mlflow_configs


def create_maia_namespace_values(namespace_config, cluster_config, config_folder, minio_configs=None, mlflow_configs=None):
    """
    Create MAIA namespace values for deployment.

    Parameters
    ----------
    namespace_config : dict
        Configuration for the namespace, including group ID and users.
    cluster_config : dict
        Configuration for the cluster, including SSH port type, port range, and storage class.
    config_folder : str
        Path to the folder where configuration files will be saved.
    minio_configs : dict, optional
        Configuration for MinIO, including access keys and console keys. Defaults to None.
    mlflow_configs : dict, optional
        Configuration for MLflow, including user and password. Defaults to None.

    Returns
    -------
    dict
        A dictionary containing the namespace, release name, chart name, repository URL, chart version,
        and the path to the generated values file.
    """

    maia_metallb_ip = cluster_config.get("maia_metallb_ip", None)
    ssh_ports = get_ssh_ports(
        len(namespace_config["users"]) + 1,
        cluster_config["ssh_port_type"],
        cluster_config["port_range"],
        maia_metallb_ip=maia_metallb_ip,
    )
    ssh_port_list = get_ssh_port_dict(
        cluster_config["ssh_port_type"],
        namespace_config["group_ID"].lower().replace("_", "-"),
        cluster_config["port_range"],
        maia_metallb_ip=maia_metallb_ip,
    )

    ssh_port_dict = {list(entry.keys())[0]: list(entry.values())[0] for entry in ssh_port_list}

    users = []

    if cluster_config["ssh_port_type"] == "LoadBalancer":
        for user in namespace_config["users"]:
            if "jupyter-" + convert_username_to_jupyterhub_username(user) in ssh_port_dict:
                users.append(
                    {
                        "jupyterhub_username": convert_username_to_jupyterhub_username(user),
                        "sshPort": ssh_port_dict["jupyter-" + convert_username_to_jupyterhub_username(user)],
                    }
                )
            else:
                users.append({"jupyterhub_username": convert_username_to_jupyterhub_username(user), "sshPort": ssh_ports.pop(0)})
    else:
        for ssh_port, user in zip(ssh_ports[:-1], namespace_config["users"]):
            if "jupyter-" + convert_username_to_jupyterhub_username(user) in ssh_port_dict:
                users.append(
                    {
                        "jupyterhub_username": convert_username_to_jupyterhub_username(user),
                        "sshPort": ssh_port_dict["jupyter-" + convert_username_to_jupyterhub_username(user)],
                    }
                )
            else:
                users.append({"jupyterhub_username": convert_username_to_jupyterhub_username(user), "sshPort": ssh_port})

    namespace = namespace_config["group_ID"].lower().replace("_", "-")

    if cluster_config["ssh_port_type"] == "LoadBalancer":
        if f"{namespace}-orthanc-svc-orthanc" in ssh_port_dict:
            orthanc_ssh_port = ssh_port_dict[f"{namespace}-orthanc-svc-orthanc"]
        else:
            orthanc_ssh_port = ssh_ports.pop(0)
    else:
        if f"{namespace}-orthanc-svc-orthanc" in ssh_port_dict:
            orthanc_ssh_port = ssh_port_dict[f"{namespace}-orthanc-svc-orthanc"]
        else:
            orthanc_ssh_port = ssh_ports[-1]

    minimal_deployment = False
    if minio_configs is None or mlflow_configs is None:
        minimal_deployment = True

    maia_namespace_values = {
        "pvc": {"pvc_type": cluster_config["shared_storage_class"], "access_mode": "ReadWriteMany", "size": "10Gi"},
        "chart_name": "maia-namespace",
        "chart_version": "1.7.2",
        "repo_url": os.environ.get("MAIA_PRIVATE_REGISTRY", None),
        "namespace": namespace_config["group_ID"].lower().replace("_", "-"),
        "serviceType": cluster_config["ssh_port_type"],
        "users": users,
        "orthanc": {"port": orthanc_ssh_port},
        "metallbSharedIp": cluster_config.get("metallb_shared_ip", False),
        "metallbIpPool": cluster_config.get("metallb_ip_pool", False),
        "loadBalancerIp": cluster_config.get("maia_metallb_ip", False),
    }
    if minimal_deployment:
        maia_namespace_values["chart_name"] = "maia-namespace"
        maia_namespace_values["chart_version"] = "1.7.2"
        maia_namespace_values["repo_url"] = "https://kthcloud.github.io/MAIA/"

    if "imagePullSecrets" in cluster_config:
        maia_namespace_values["dockerRegistrySecret"] = {
            "enabled": True,
            "dockerRegistrySecretName": cluster_config["imagePullSecrets"],
            "dockerRegistrySecret": encode_docker_registry_secret(
                cluster_config["docker_server"], cluster_config["docker_username"], cluster_config["docker_password"]
            ),
        }

    if minio_configs:
        maia_namespace_values["minio"] = {
            "enabled": True,
            "consoleDomain": "https://{}.{}/minio-console".format(namespace_config["group_subdomain"], cluster_config["domain"]),
            "namespace": namespace_config["group_ID"].lower().replace("_", "-"),
            "storageClassName": cluster_config["storage_class"],
            "storageSize": "10Gi",
            "accessKey": minio_configs["access_key"],
            "secretKey": minio_configs["secret_key"],
            "clientId": cluster_config["keycloak"]["client_id"],
            "clientSecret": cluster_config["keycloak"]["client_secret"],
            "openIdConfigUrl": cluster_config["keycloak"]["issuer_url"] + "/.well-known/openid-configuration",
            "consoleAccessKey": minio_configs["console_access_key"],
            "consoleSecretKey": minio_configs["console_secret_key"],
            "ingress": {
                "annotations": {},
                "host": "{}.{}".format(namespace_config["group_subdomain"], cluster_config["domain"]),
                "path": "minio-console",
                "serviceName": f"{namespace}-mlflow-mkg",
            },
        }

        if "nginx_cluster_issuer" in cluster_config:
            maia_namespace_values["minio"]["ingress"]["annotations"]["cert-manager.io/cluster-issuer"] = cluster_config[
                "nginx_cluster_issuer"
            ]
            maia_namespace_values["minio"]["ingress"]["annotations"]["nginx.ingress.kubernetes.io/proxy-body-size"] = "10g"
            maia_namespace_values["minio"]["ingress"]["tlsSecretName"] = "{}.{}-tls".format(
                namespace_config["group_subdomain"], cluster_config["domain"]
            )
        if "traefik_resolver" in cluster_config:
            maia_namespace_values["minio"]["ingress"]["annotations"][
                "traefik.ingress.kubernetes.io/router.entrypoints"
            ] = "websecure"
            maia_namespace_values["minio"]["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.tls"] = "true"
            maia_namespace_values["minio"]["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.tls.certresolver"] = (
                cluster_config["traefik_resolver"]
            )
        if cluster_config["url_type"] == "subpath":
            maia_namespace_values["minio"]["consoleDomain"] = "https://{}/{}-minio-console".format(
                cluster_config["domain"], namespace_config["group_ID"].lower().replace("_", "-")
            )
            maia_namespace_values["minio"]["ingress"]["host"] = "{}".format(cluster_config["domain"])
            maia_namespace_values["minio"]["ingress"]["path"] = "{}-minio-console".format(
                namespace_config["group_ID"].lower().replace("_", "-")
            )

    if mlflow_configs:
        maia_namespace_values["mlflow"] = {
            "enabled": True,
            "user": base64.b64decode(mlflow_configs["mlflow_user"]).decode("ascii"),
            "password": base64.b64decode(mlflow_configs["mlflow_password"]).decode("ascii"),
        }

    enable_cifs = namespace_config.get("extra_configs", {}).get("enable_cifs", False)
    if enable_cifs:
        maia_namespace_values["cifs"] = {"enabled": True, "encryption": {"publicKey": ""}}  # base64 encoded}
    namespace_id = namespace_config["group_ID"].lower().replace("_", "-")
    Path(config_folder).joinpath(namespace_config["group_ID"], "maia_namespace_values").mkdir(parents=True, exist_ok=True)
    with open(
        Path(config_folder).joinpath(namespace_config["group_ID"], "maia_namespace_values", "namespace_values.yaml"), "w"
    ) as f:
        f.write(OmegaConf.to_yaml(maia_namespace_values))

    return {
        "namespace": maia_namespace_values["namespace"],
        "release": f"{namespace_id}-namespace",
        "chart": maia_namespace_values["chart_name"],
        "repo": maia_namespace_values["repo_url"],
        "version": maia_namespace_values["chart_version"],
        "values": str(
            Path(config_folder).joinpath(namespace_config["group_ID"], "maia_namespace_values", "namespace_values.yaml")
        ),
    }


def create_filebrowser_values(namespace_config, cluster_config, config_folder, mlflow_configs=None):
    """
    Create and write configuration values for deploying the MAIA Filebrowser Helm chart.
    This function generates a dictionary of configuration values required to deploy the MAIA Filebrowser
    application in a Kubernetes namespace. It handles image configuration, environment variables, volume
    mounts, CIFS volume setup, and ingress settings for both NGINX and Traefik ingress controllers. The
    resulting configuration is written to a YAML file in the specified config folder.
    
    Parameters
    ----------
    namespace_config : dict
        Dictionary containing namespace-specific configuration, including group ID, subdomain, and users.
    cluster_config : dict
        Dictionary containing cluster-specific configuration, such as docker server, image pull secrets,
        domain, and optional ingress settings.
    config_folder : str or Path
        Path to the folder where the generated configuration YAML file will be saved.
    mlflow_configs : dict, optional
        Optional dictionary containing MLflow configuration, specifically the base64-encoded
        'mlflow_password'. If not provided, a new human-memorable password is generated.
    
    Returns
    -------
    dict
        A dictionary containing:
            - 'namespace': The Kubernetes namespace for deployment.
            - 'release': The Helm release name.
            - 'chart': The Helm chart name.
            - 'repo': The Helm chart repository URL.
            - 'version': The Helm chart version.
            - 'values': Path to the generated YAML values file.
    
    Notes
    -----
    - The function expects certain helper functions and environment variables to be available, such as
      `generate_human_memorable_password`, `convert_username_to_jupyterhub_username`, and `OmegaConf`.
    - The CIFS server address is read from the 'CIFS_SERVER' environment variable.
    """

    namespace_id = namespace_config["group_ID"].lower().replace("_", "-")

    maia_filebrowser_values = {
        "chart_name": "maia-filebrowser",
        "chart_version": "1.0.0",
        "repo_url": "https://kthcloud.github.io/MAIA/",
        "namespace": namespace_config["group_ID"].lower().replace("_", "-"),
    }

    maia_filebrowser_values["image"] = {
        "repository": cluster_config["docker_server"] + "/maia/maia-filebrowser",
        "tag": "1.0",
    }

    maia_filebrowser_values["imagePullSecrets"] = [{"name": cluster_config["imagePullSecrets"]}]
    if mlflow_configs is None:
        pw = generate_human_memorable_password(16)
    else:
        pw = base64.b64decode(mlflow_configs["mlflow_password"]).decode("ascii")
    maia_filebrowser_values["env"] = [
        {"name": "RUN_FILEBROWSER", "value": "True"},
        {"name": "n_users", "value": "1"},
        {"name": "user", "value": "maia-admin"},
        {"name": "password", "value": pw},
    ]

    maia_filebrowser_values["volumeMounts"] = [{"name": "cifs", "mountPath": "/home/cifs"}]

    cifs_user = convert_username_to_jupyterhub_username(namespace_config["users"][0])
    maia_filebrowser_values["volumes"] = [
        {
            "name": "cifs",
            "flexVolume": {
                "driver": "fstab/cifs",
                "fsType": "cifs",
                "secretRef": {"name": cifs_user + "-cifs"},
                "options": {
                    "mountOptions": "dir_mode=0777,file_mode=0777,iocharset=utf8,noperm,nounix,rw",
                    "networkPath": os.environ.get("CIFS_SERVER", "N/A"),
                },
            },
        }
    ]
        
    maia_filebrowser_values["ingress"] = {
        "enabled": True,
        "annotations": {},
        "hosts": [{"host": "cifs.{}.{}".format(namespace_config["group_subdomain"], cluster_config["domain"]),
                   "paths": [{"path": "/", "pathType": "ImplementationSpecific"}]}],
        "tls": [{"hosts": ["cifs.{}.{}".format(namespace_config["group_subdomain"], cluster_config["domain"])]}],
    }
    if "nginx_cluster_issuer" in cluster_config:
        maia_filebrowser_values["ingress"]["annotations"]["cert-manager.io/cluster-issuer"] = cluster_config["nginx_cluster_issuer"]
        maia_filebrowser_values["ingress"]["annotations"]["nginx.ingress.kubernetes.io/proxy-body-size"] = "10g"
        maia_filebrowser_values["ingress"]["tls"][0]["secretName"] = "{}.{}-tls".format(namespace_config["group_subdomain"], cluster_config["domain"])
    if "traefik_resolver" in cluster_config:
        maia_filebrowser_values["ingress"]["annotations"][
            "traefik.ingress.kubernetes.io/router.entrypoints"
        ] = "websecure"
        maia_filebrowser_values["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.tls"] = "true"
        maia_filebrowser_values["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.tls.certresolver"] = (
            cluster_config["traefik_resolver"]
        )

    Path(config_folder).joinpath(namespace_config["group_ID"], "maia_filebrowser_values").mkdir(parents=True, exist_ok=True)
    with open(Path(config_folder).joinpath(namespace_config["group_ID"], "maia_filebrowser_values", "maia_filebrowser_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(maia_filebrowser_values))

    return {
        "namespace": maia_filebrowser_values["namespace"],
        "release": f"{namespace_id}-namespace",
        "chart": maia_filebrowser_values["chart_name"],
        "repo": maia_filebrowser_values["repo_url"],
        "version": maia_filebrowser_values["chart_version"],
        "values": str(
            Path(config_folder).joinpath(namespace_config["group_ID"], "maia_filebrowser_values", "maia_filebrowser_values.yaml")
        ),
    }


async def get_maia_toolkit_apps(group_id, token, argo_cd_host):
    """
    Retrieve and print information about a specific project and its associated applications from Argo CD.

    Parameters
    ----------
    group_id : str
        The group identifier used to construct project and application names.
    token : str
        The authorization token for accessing the Argo CD API.
    argo_cd_host : str
        The host URL of the Argo CD server.

    Returns
    -------
    None

    Raises
    ------
    ApiException
        If there is an error when calling the Argo CD API.
    """
    configuration = argocd_client.Configuration(host=argo_cd_host)
    with argocd_client.ApiClient(configuration) as api_client:
        api_client.default_headers["Authorization"] = f"Bearer {token}"
        api_instance = argocd_client.ProjectServiceApi(api_client)
        api_instance_apps = argocd_client.ApplicationServiceApi(api_client)
        name = group_id.lower().replace("_", "-")

        try:
            api_response = api_instance.get_mixin6(name)
            pprint(api_response)
        except ApiException as e:
            print("Exception when calling ProjectServiceApi->get_mixin6: %s\n" % e)

        names = [
            f'{group_id.lower().replace("_", "-")}-namespace',
            f'{group_id.lower().replace("_", "-")}-jupyterhub',
            f'{group_id.lower().replace("_", "-")}-oauth2-proxy',
        ]
        project = [group_id.lower().replace("_", "-")]
        for name in names:
            try:
                api_response = api_instance_apps.get_mixin9(name, project=project)
                pprint(api_response)
            except ApiException as e:
                print("Exception when calling ApplicationServiceApi->get_mixin9: %s\n" % e)


async def install_maia_project(
    group_id, values_file, argo_cd_namespace, project_chart, project_repo=None, project_version=None, json_key_path=None
):
    """
    Installs or upgrades a MAIA project using the specified Helm chart and values file.

    Parameters
    ----------
    group_id : str
        The group ID for the project. This will be used as the release name.
    values_file : str
        Path to the YAML file containing the values for the Helm chart.
    argo_cd_namespace : str
        The namespace in which to install the project.
    project_chart : str
        The name of the Helm chart to use for the project.
    project_repo : str, optional
        The repository URL where the Helm chart is located. Defaults to None.
    project_version : str, optional
        The version of the Helm chart to use. Defaults to None.
    json_key_path : str, optional
        Path to the JSON key file for authentication with the Helm registry. Defaults to None.

    Returns
    -------
    None

    Raises
    ------
    FileNotFoundError
        If the values file does not exist.
    yaml.YAMLError
        If there is an error parsing the values file.
    Exception
        If there is an error during the installation or upgrade process.

    Example
    -------
    await install_maia_project(
        group_id="example_group",
        values_file="/path/to/values.yaml",
        argo_cd_namespace="default",
        project_chart="example_chart",
        project_repo="https://example.com/charts",
        project_version="1.0.0"
    )
    """
    client = Client(kubeconfig=os.environ["KUBECONFIG"])

    if not project_repo.startswith("http"):
        chart = str("/tmp/" + project_chart + "-" + project_version + ".tgz")
        project_chart = "oci://" + project_repo + "/" + project_chart

        try:
            try:
                with open(json_key_path, "r") as f:
                    docker_credentials = json.load(f)
                    username = docker_credentials.get("harbor_username")
                    password = docker_credentials.get("harbor_password")
            except Exception:
                with open(json_key_path, "r") as f:
                    docker_credentials = f.read()
                    username = "_json_key"
                    password = docker_credentials
            result = subprocess.run(
                ["helm", "registry", "login", project_repo, "-u", username, "--password-stdin"],
                input=password.encode(),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True,
            )
            print("✅ Helm registry login successful.")
            print(result.stdout.decode())
        except subprocess.CalledProcessError as e:
            print("❌ Helm registry login failed.")
            print("STDOUT:", e.stdout.decode())
            print("STDERR:", e.stderr.decode())
            await asyncio.sleep(1)
            return "Deployment failed: Helm registry login failed."
        subprocess.run(["helm", "pull", project_chart, "-d", "/tmp", "--version", project_version], check=True)

        subprocess.run(
            [
                "helm",
                "upgrade",
                "--install",
                group_id.lower().replace("_", "-"),
                chart,
                "--namespace",
                argo_cd_namespace,
                "--values",
                str(values_file),
                "--wait",
            ],
            check=True,
        )
        await asyncio.sleep(1)
        return ""

    chart = await client.get_chart(project_chart, repo=project_repo, version=project_version)

    with open(values_file) as f:
        values = yaml.safe_load(f)

    revision = await client.install_or_upgrade_release(
        group_id.lower().replace("_", "-"), chart, values, namespace=argo_cd_namespace, wait=True
    )
    print(revision.release.name, revision.release.namespace, revision.revision, str(revision.status))

    return ""


def create_maia_admin_toolkit_values(config_folder, project_id, cluster_config_dict, maia_config_dict):
    """
    Creates and writes the MAIA admin toolkit values to a YAML file.

    Parameters
    ----------
    config_folder : str
        The path to the configuration folder.
    project_id : str
        The project identifier.
    cluster_config_dict : dict
        Dictionary containing cluster configuration values.
    maia_config_dict : dict
        Dictionary containing MAIA configuration values.

    Returns
    -------
    dict
        A dictionary containing the namespace, release name, chart name, repository URL, chart version,
        and the path to the generated values YAML file.
    """
    admin_group_id = maia_config_dict["admin_group_ID"]

    admin_toolkit_values = {
        "namespace": "maia-admin-toolkit",
        "repo_url": "https://kthcloud.github.io/MAIA/",
        "chart_name": "maia-admin-toolkit",
        "chart_version": "1.2.6",
    }

    admin_toolkit_values.update(
        {
            "argocd": {
                "enabled": True,
                "argocd_namespace": "argocd",
                "argocd_domain": "argocd." + cluster_config_dict["domain"],
                "keycloak_issuer_url": "https://iam." + cluster_config_dict["domain"] + "/realms/maia",
                "keycloak_client_id": "maia",
                "keycloak_client_secret": cluster_config_dict["keycloak"]["client_secret"],
            },
            "certResolver": cluster_config_dict["traefik_resolver"],
            "admin_group_ID": admin_group_id,
            "harbor": {
                "enabled": True,
                "values": {"namespace": "harbor", "storageClassName": cluster_config_dict["storage_class"]},
            },
            "dashboard": {"enabled": True, "dashboard_domain": "dashboard." + cluster_config_dict["domain"]},
        }
    )

    Path(config_folder).joinpath(project_id, "maia_admin_toolkit_values").mkdir(parents=True, exist_ok=True)

    with open(Path(config_folder).joinpath(project_id, "maia_admin_toolkit_values", "maia_admin_toolkit_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(admin_toolkit_values))

    return {
        "namespace": admin_toolkit_values["namespace"],
        "release": f"{project_id}-toolkit",
        "chart": admin_toolkit_values["chart_name"],
        "repo": admin_toolkit_values["repo_url"],
        "version": admin_toolkit_values["chart_version"],
        "values": str(Path(config_folder).joinpath(project_id, "maia_admin_toolkit_values", "maia_admin_toolkit_values.yaml")),
    }


def create_harbor_values(config_folder, project_id, cluster_config_dict):
    """
    Create and save Harbor values configuration for a given project and cluster configuration.

    Parameters
    ----------
    config_folder : str
        The path to the configuration folder where the Harbor values file will be saved.
    project_id : str
        The unique identifier for the project.
    cluster_config_dict : dict
        A dictionary containing cluster configuration details, including:
            - domain (str): The domain name for the Harbor registry.
            - ingress_class (str): The ingress class to be used (e.g., "maia-core-traefik", "nginx").
            - traefik_resolver (str, optional): The Traefik resolver to be used if ingress_class is "maia-core-traefik".

    Returns
    -------
    dict
        A dictionary containing the following keys:
        - namespace (str): The Kubernetes namespace for Harbor.
        - release (str): The release name for the Harbor Helm chart.
        - chart (str): The name of the Harbor Helm chart.
        - repo (str): The URL of the Harbor Helm chart repository.
        - version (str): The version of the Harbor Helm chart.
        - values (str): The path to the generated Harbor values YAML file.
    """
    domain = cluster_config_dict["domain"]
    harbor_values = {
        "namespace": "harbor",
        "repo_url": "https://helm.goharbor.io",
        "chart_name": "harbor",
        "chart_version": "1.16.0",
    }

    harbor_values.update(
        {
            "expose": {
                "type": "ingress",
                "tls": {"enabled": True},
                "ingress": {
                    "hosts": {"core": f"registry.{domain}"},
                    "annotations": {},
                    "controller": "default",
                    "className": cluster_config_dict["ingress_class"],
                },
            },
            "externalURL": f"https://registry.{domain}",
            "persistence": {
                "enabled": True,
                "resourcePolicy": "keep",
                "persistentVolumeClaim": {
                    "registry": {
                        "existingClaim": "pvc-harbor",
                        "subPath": "registry",
                        "storageClass": cluster_config_dict["storage_class"],
                        "accessMode": "ReadWriteMany",
                    },
                    "jobservice": {
                        "jobLog": {
                            "existingClaim": "pvc-harbor",
                            "subPath": "job_logs",
                            "storageClass": cluster_config_dict["storage_class"],
                            "accessMode": "ReadWriteMany",
                        }
                    },
                    "database": {
                        "existingClaim": "pvc-harbor",
                        "subPath": "database",
                        "storageClass": cluster_config_dict["storage_class"],
                        "accessMode": "ReadWriteMany",
                    },
                    "redis": {
                        "existingClaim": "pvc-harbor",
                        "subPath": "redis",
                        "storageClass": cluster_config_dict["storage_class"],
                        "accessMode": "ReadWriteMany",
                    },
                    "trivy": {
                        "existingClaim": "pvc-harbor",
                        "subPath": "trivy",
                        "storageClass": cluster_config_dict["storage_class"],
                        "accessMode": "ReadWriteMany",
                    },
                },
                "imageChartStorage": {"type": "filesystem"},
            },
            "database": {"internal": {"password": "harbor"}},
            "metrics": {
                "enabled": True,
                "core": {"path": "/metrics", "port": 8001},
                "registry": {"path": "/metrics", "port": 8001},
                "jobservice": {"path": "/metrics", "port": 8001},
                "exporter": {"path": "/metrics", "port": 8001},
            },
        }
    )

    if cluster_config_dict["ingress_class"] == "maia-core-traefik":
        harbor_values["expose"]["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.entrypoints"] = "websecure"
        harbor_values["expose"]["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.tls"] = "true"
        harbor_values["expose"]["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.tls.certresolver"] = (
            cluster_config_dict["traefik_resolver"]
        )
    elif cluster_config_dict["ingress_class"] == "nginx":
        harbor_values["expose"]["ingress"]["annotations"]["cert-manager.io/cluster-issuer"] = "cluster-issuer"

    Path(config_folder).joinpath(project_id, "harbor_values").mkdir(parents=True, exist_ok=True)
    with open(Path(config_folder).joinpath(project_id, "harbor_values", "harbor_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(harbor_values))

    return {
        "namespace": harbor_values["namespace"],
        "release": f"{project_id}-harbor",
        "chart": harbor_values["chart_name"],
        "repo": harbor_values["repo_url"],
        "version": harbor_values["chart_version"],
        "values": str(Path(config_folder).joinpath(project_id, "harbor_values", "harbor_values.yaml")),
    }


def create_keycloak_values(config_folder, project_id, cluster_config_dict):
    """
    Generates Keycloak Helm chart values and writes them to a YAML file.

    Parameters
    ----------
    config_folder : str
        The path to the configuration folder where the YAML file will be saved.
    project_id : str
        The project identifier used to create a unique namespace and release name.
    cluster_config_dict : dict
        A dictionary containing cluster configuration details such as domain, ingress class, and traefik resolver.

    Returns
    -------
    dict
        A dictionary containing the namespace, release name, chart name, repository URL, chart version,
        and the path to the generated values YAML file.
    """
    keycloak_values = {
        "namespace": "keycloak",
        "repo_url": "https://charts.bitnami.com/bitnami",
        "chart_name": "keycloak",
        "chart_version": "24.2.0",
    }

    keycloak_values.update(
        {
            "extraEnvVars": [
                {"name": "KEYCLOAK_EXTRA_ARGS", "value": "--import-realm"},
                {"name": "PROXY_ADDRESS_FORWARDING", "value": "true"},
                {"name": "KEYCLOAK_HOSTNAME", "value": "iam." + cluster_config_dict["domain"]},
            ],
            "proxy": "edge",
            "ingress": {
                "enabled": True,
                "tls": True,
                "ingressClassName": cluster_config_dict["ingress_class"],
                "hostname": "iam." + cluster_config_dict["domain"],
                "annotations": {},
            },
            "extraVolumeMounts": [{"name": "keycloak-import", "mountPath": "/opt/bitnami/keycloak/data/import"}],
            "extraVolumes": [{"name": "keycloak-import", "configMap": {"name": "maia-realm-import"}}],
        }
    )

    if cluster_config_dict["ingress_class"] == "maia-core-traefik":
        keycloak_values["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.entrypoints"] = "websecure"
        keycloak_values["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.tls"] = "true"
        keycloak_values["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.tls.certresolver"] = cluster_config_dict[
            "traefik_resolver"
        ]
    elif cluster_config_dict["ingress_class"] == "nginx":
        keycloak_values["ingress"]["annotations"]["cert-manager.io/cluster-issuer"] = "cluster-issuer"

    Path(config_folder).joinpath(project_id, "keycloak_values").mkdir(parents=True, exist_ok=True)
    with open(Path(config_folder).joinpath(project_id, "keycloak_values", "keycloak_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(keycloak_values))

    return {
        "namespace": keycloak_values["namespace"],
        "release": f"{project_id}-keycloak",
        "chart": keycloak_values["chart_name"],
        "repo": keycloak_values["repo_url"],
        "version": keycloak_values["chart_version"],
        "values": str(Path(config_folder).joinpath(project_id, "keycloak_values", "keycloak_values.yaml")),
    }


def create_loginapp_values(config_folder, project_id, cluster_config_dict):
    """
    Creates and writes the loginapp values configuration file for a given project and cluster configuration.

    Parameters
    ----------
    config_folder : str
        The base directory where the configuration files will be stored.
    project_id : str
        The unique identifier for the project.
    cluster_config_dict : dict
        A dictionary containing cluster configuration details, including:
            - keycloak.client_secret (str): The client secret for Keycloak.
            - domain (str): The domain name for the cluster.
            - ingress_class (str): The ingress class to be used (e.g., "maia-core-traefik" or "nginx").
            - traefik_resolver (str, optional): The Traefik resolver to be used if ingress_class is "maia-core-traefik".

    Returns
    -------
    dict
        A dictionary containing the namespace, release name, chart name, repository URL, chart version,
        and the path to the generated values file.

    Raises
    ------
    KeyError
        If required keys are missing from the cluster_config_dict.
    OSError
        If there is an error creating directories or writing the configuration file.
    """
    loginapp_values = {
        "namespace": "authentication",
        "repo_url": "https://storage.googleapis.com/loginapp-releases/charts/",
        "chart_name": "loginapp",
        "chart_version": "1.3.0",
    }

    secret = token_urlsafe(16).replace("-", "_")
    client_id = "maia"
    client_secret = cluster_config_dict["keycloak"]["client_secret"]
    issuer_url = "https://iam." + cluster_config_dict["domain"] + "/realms/maia"
    cluster_server_address = "https://" + cluster_config_dict["domain"] + ":16443"
    ca_file = "/var/snap/microk8s/current/certs/ca.crt"

    loginapp_values.update(
        {
            "env": {"LOGINAPP_NAME": "MAIA Login"},
            "configOverwrites": {"oidc": {"scopes": ["openid", "profile", "email"]}, "service": {"type": "ClusterIP"}},
            "ingress": {
                "enabled": True,
                "annotations": {},
                "tls": [{"hosts": ["login." + cluster_config_dict["domain"]]}],
                "hosts": [{"host": "login." + cluster_config_dict["domain"], "paths": [{"path": "/", "pathType": "Prefix"}]}],
            },
            "config": {
                "tls": {"enabled": False},
                "issuerInsecureSkipVerify": True,
                "refreshToken": True,
                "clientRedirectURL": "https://login." + cluster_config_dict["domain"] + "/callback",
                "secret": secret,
                "clientID": client_id,
                "clientSecret": client_secret,
                "issuerURL": issuer_url,
                "clusters": [
                    {
                        "server": cluster_server_address,
                        "name": "MAIA",
                        "insecure-skip-tls-verify": True,
                        "certificate-authority": ca_file,
                    }
                ],
            },
        }
    )

    if cluster_config_dict["ingress_class"] == "maia-core-traefik":
        loginapp_values["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.entrypoints"] = "websecure"
        loginapp_values["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.tls"] = "true"
        loginapp_values["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.tls.certresolver"] = cluster_config_dict[
            "traefik_resolver"
        ]
    elif cluster_config_dict["ingress_class"] == "nginx":
        loginapp_values["ingress"]["annotations"]["cert-manager.io/cluster-issuer"] = "cluster-issuer"
        loginapp_values["ingress"]["tls"][0]["secretName"] = "loginapp." + cluster_config_dict["domain"]

    Path(config_folder).joinpath(project_id, "loginapp_values").mkdir(parents=True, exist_ok=True)
    with open(Path(config_folder).joinpath(project_id, "loginapp_values", "loginapp_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(loginapp_values))

    return {
        "namespace": loginapp_values["namespace"],
        "release": f"{project_id}-loginapp",
        "chart": loginapp_values["chart_name"],
        "repo": loginapp_values["repo_url"],
        "version": loginapp_values["chart_version"],
        "values": str(Path(config_folder).joinpath(project_id, "loginapp_values", "loginapp_values.yaml")),
    }


def create_minio_operator_values(config_folder, project_id, cluster_config_dict):
    """
    Creates and writes MinIO operator values to a YAML file and returns a dictionary with deployment details.

    Parameters
    ----------
    config_folder : str
        The path to the configuration folder.
    project_id : str
        The unique identifier for the project.
    cluster_config_dict : dict
        A dictionary containing cluster configuration details.

    Returns
    -------
    dict
        A dictionary containing the namespace, release name, chart name, repository URL, chart version,
        and the path to the generated YAML values file.
    """
    minio_operator_values = {
        "namespace": "minio-operator",
        "repo_url": "https://operator.min.io",
        "chart_name": "operator",
        "chart_version": "6.0.4",
    }

    Path(config_folder).joinpath(project_id, "minio_operator_values").mkdir(parents=True, exist_ok=True)
    with open(Path(config_folder).joinpath(project_id, "minio_operator_values", "minio_operator_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(minio_operator_values))

    return {
        "namespace": minio_operator_values["namespace"],
        "release": f"{project_id}-minio-operator",
        "chart": minio_operator_values["chart_name"],
        "repo": minio_operator_values["repo_url"],
        "version": minio_operator_values["chart_version"],
        "values": str(Path(config_folder).joinpath(project_id, "minio_operator_values", "minio_operator_values.yaml")),
    }


def create_maia_dashboard_values(config_folder, project_id, cluster_config_dict, maia_config_dict):
    """
    Create MAIA dashboard values for Helm chart deployment.

    Parameters
    ----------
    config_folder : str
        The path to the configuration folder.
    project_id : str
        The project identifier.
    cluster_config_dict : dict
        Dictionary containing cluster configuration details.
    maia_config_dict : dict
        Dictionary containing MAIA configuration details.

    Returns
    -------
    dict
        A dictionary containing the namespace, release name, chart name, repository URL, chart version,
        and the path to the generated values YAML file.
    """

    maia_dashboard_values = {
        "namespace": "maia-dashboard",
        "repo_url": "https://kthcloud.github.io/MAIA/",
        "chart_name": "maia-dashboard",
        "chart_version": "0.1.6",
    }

    maia_dashboard_values.update(
        {
            "image": {
                "repository": maia_config_dict["dashboard_image"],
                "pullPolicy": "IfNotPresent",
                "tag": maia_config_dict["dashboard_version"],
            },
            "imagePullSecrets": [{"name": "registry." + cluster_config_dict["domain"]}],  # TODO: Update
            "storageClass": cluster_config_dict["storage_class"],
            "ingress": {
                "enabled": True,
                "className": cluster_config_dict["ingress_class"],
                "annotations": {},
                "hosts": [
                    {
                        "host": cluster_config_dict["domain"],
                        "paths": [
                            {"path": "/maia/", "pathType": "Prefix"},
                            {"path": "/maia-api/", "pathType": "Prefix"},
                            {"path": "/", "pathType": "Prefix"},
                        ],
                    }
                ],
                "tls": [{"hosts": [cluster_config_dict["domain"]]}],
            },
            "gpuList": maia_config_dict["gpu_list"],
            "dashboard": {
                "host": cluster_config_dict["domain"],
                "api_secret_key": maia_config_dict["dashboard_api_secret"],
                "keycloak": {
                    "client_id": "maia",
                    "client_secret": cluster_config_dict["keycloak"]["client_secret"],
                    "url": "https://iam." + cluster_config_dict["domain"] + "/",
                    "realm": "maia",
                    "username": "admin",
                },
                "argocd_server": "https://argocd." + cluster_config_dict["domain"],
                "argocd_cluster_name": cluster_config_dict["cluster_name"],
                "local_db_path": "/etc/MAIA-Dashboard/db",
            },
            "argocd_namespace": maia_config_dict["argocd_namespace"],
            "admin_group_ID": maia_config_dict["admin_group_ID"],
            "core_project_chart": maia_config_dict["core_project_chart"],
            "core_project_repo": maia_config_dict["core_project_repo"],
            "core_project_version": maia_config_dict["core_project_version"],
            "admin_project_chart": maia_config_dict["admin_project_chart"],
            "admin_project_repo": maia_config_dict["admin_project_repo"],
            "admin_project_version": maia_config_dict["admin_project_version"],
            "maia_project_chart": maia_config_dict["maia_project_chart"],
            "maia_project_repo": maia_config_dict["maia_project_repo"],
            "maia_project_version": maia_config_dict["maia_project_version"],
            "maia_pro_project_chart": maia_config_dict["maia_pro_project_chart"],
            "maia_pro_project_repo": maia_config_dict["maia_pro_project_repo"],
            "maia_pro_project_version": maia_config_dict["maia_pro_project_version"],
            "maia_workspace_version": maia_config_dict["maia_workspace_version"],
            "maia_workspace_image": maia_config_dict["maia_workspace_image"],
            "maia_workspace_pro_version": maia_config_dict["maia_workspace_pro_version"],
            "maia_workspace_pro_image": maia_config_dict["maia_workspace_pro_image"],
            "maia_orthanc_version": maia_config_dict["maia_orthanc_version"],
            "maia_orthanc_image": maia_config_dict["maia_orthanc_image"],
            "maia_monai_toolkit_image": maia_config_dict["maia_monai_toolkit_image"],
            "name": "maia-dashboard",
            "dockerRegistrySecretName": "registry." + cluster_config_dict["domain"],
            "dockerRegistryUsername": cluster_config_dict["docker_username"],
            "dockerRegistryPassword": cluster_config_dict["docker_password"],
            "dockerRegistryEmail": cluster_config_dict["docker_email"],
            "dockerRegistryServer": "registry." + cluster_config_dict["domain"],
        }
    )

    if cluster_config_dict["ingress_class"] == "maia-core-traefik":
        maia_dashboard_values["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.entrypoints"] = "websecure"
        maia_dashboard_values["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.tls"] = "true"
        maia_dashboard_values["ingress"]["annotations"]["traefik.ingress.kubernetes.io/router.tls.certresolver"] = (
            cluster_config_dict["traefik_resolver"]
        )
    elif cluster_config_dict["ingress_class"] == "nginx":
        maia_dashboard_values["ingress"]["annotations"]["nginx.ingress.kubernetes.io/proxy-body-size"] = "8g"
        maia_dashboard_values["ingress"]["annotations"]["nginx.ingress.kubernetes.io/proxy-read-timeout"] = "300"
        maia_dashboard_values["ingress"]["annotations"]["nginx.ingress.kubernetes.io/proxy-send-timeout"] = "300"
        maia_dashboard_values["ingress"]["annotations"]["cert-manager.io/cluster-issuer"] = "cluster-issuer"
        maia_dashboard_values["ingress"]["tls"][0]["secretName"] = cluster_config_dict["domain"]

    maia_dashboard_values["clusters"] = [cluster_config_dict]

    if "discord_url" in maia_config_dict:
        maia_dashboard_values["dashboard"]["discord_url"] = maia_config_dict["discord_url"]
    if "discord_signup_url" in maia_config_dict:
        maia_dashboard_values["dashboard"]["discord_signup_url"] = maia_config_dict["discord_signup_url"]
    if "discord_support_url" in maia_config_dict:
        maia_dashboard_values["dashboard"]["discord_support_url"] = maia_config_dict["discord_support_url"]

    debug = False

    if debug:

        maia_dashboard_values["env"] = [
            {"name": "DEBUG", "value": "True"},
            {"name": "CLUSTER_CONFIG_PATH", "value": "/etc/MAIA-Dashboard/config"},
            {"name": "CONFIG_PATH", "value": "/etc/MAIA-Dashboard/config"},
            {"name": "MAIA_CONFIG_PATH", "value": "/etc/MAIA-Dashboard/config/maia_config.yaml"},
            {"name": "GLOBAL_NAMESPACES", "value": "xnat,kubeflow,istio-system"},
        ]
        maia_dashboard_values["dashboard"]["local_config_path"] = "/etc/MAIA-Dashboard/config"
    else:

        if "mysql_dashboard_password" in maia_config_dict:
            db_password = maia_config_dict["mysql_dashboard_password"]
        else:
            db_password = generate_human_memorable_password()
        maia_dashboard_values["dashboard"]["local_config_path"] = "/mnt/dashboard-config"
        cifs_server = ""
        maia_dashboard_values["env"] = [
            {"name": "DEBUG", "value": "False"},
            {"name": "CLUSTER_CONFIG_PATH", "value": "/mnt/dashboard-config"},
            {"name": "CONFIG_PATH", "value": "/mnt/dashboard-config"},
            {"name": "MAIA_CONFIG_PATH", "value": "/mnt/dashboard-config/maia_config.yaml"},
            {"name": "DB_ENGINE", "value": "mysql"},
            {"name": "DB_NAME", "value": "mysql"},
            {"name": "DB_HOST", "value": "maia-admin-maia-dashboard-mysql"},
            {"name": "DB_PORT", "value": "3306"},
            {"name": "DB_USERNAME", "value": "maia-admin"},
            {"name": "DB_PASS", "value": db_password},
            {"name": "GLOBAL_NAMESPACES", "value": "xnat,kubeflow,istio-system"},
        ]
        if maia_config_dict["cifs_server"]:
            cifs_server = maia_config_dict["cifs_server"]
            maia_dashboard_values["env"].append({"name": "CIFS_SERVER", "value": cifs_server})
        maia_dashboard_values["mysql"] = {
            "enabled": True,
            "mysqlRootPassword": db_password,
            "mysqlUser": "maia-admin",
            "mysqlPassword": db_password,
            "mysqlDatabase": "mysql",
        }

    if maia_config_dict["email_server_credentials"]:
        maia_dashboard_values["env"].extend(
            [
                {"name": "email_account", "value": maia_config_dict["email_server_credentials"]["email_account"]},
                {"name": "email_password", "value": maia_config_dict["email_server_credentials"]["email_password"]},
                {"name": "email_smtp_server", "value": maia_config_dict["email_server_credentials"]["email_smtp_server"]},
            ]
        )
    maia_dashboard_values["env"].append({"name": "MAIA_PRIVATE_REGISTRY", "value": maia_config_dict["maia_private_registry"]})
    if "minio" in maia_config_dict:
        maia_dashboard_values["dashboard"]["minio"] = {
            "url": maia_config_dict["minio"]["url"],
            "access_key": maia_config_dict["minio"]["access_key"],
            "secret_key": maia_config_dict["minio"]["secret_key"],
            "secure": maia_config_dict["minio"]["secure"],
            "bucket_name": maia_config_dict["minio"]["bucket_name"],
        }
    Path(config_folder).joinpath(project_id, "maia_dashboard_values").mkdir(parents=True, exist_ok=True)
    with open(Path(config_folder).joinpath(project_id, "maia_dashboard_values", "maia_dashboard_values.yaml"), "w") as f:
        f.write(OmegaConf.to_yaml(maia_dashboard_values))

    return {
        "namespace": maia_dashboard_values["namespace"],
        "release": f"{project_id}-dashboard",
        "chart": maia_dashboard_values["chart_name"],
        "repo": maia_dashboard_values["repo_url"],
        "version": maia_dashboard_values["chart_version"],
        "values": str(Path(config_folder).joinpath(project_id, "maia_dashboard_values", "maia_dashboard_values.yaml")),
    }



from __future__ import annotations

import asyncio
import email
import os
import smtplib
import ssl
from datetime import datetime
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from pathlib import Path

import requests
import yaml
from bs4 import BeautifulSoup
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from keycloak import KeycloakAdmin, KeycloakOpenIDConnection
from kubernetes import config
from minio import Minio
from pyhelm3 import Client

from MAIA.keycloak_utils import get_groups_in_keycloak
from MAIA.kubernetes_utils import generate_kubeconfig, get_namespaces
from MAIA_scripts.MAIA_install_project_toolkit import verify_installed_maia_toolkit


def verify_gpu_availability(global_existing_bookings, new_booking, gpu_specs):
    """
    Verify GPU availability for a new booking.

    Parameters
    ----------
    global_existing_bookings : list of dict
        A list of existing bookings where each booking is represented as a dictionary
        with keys "gpu", "start_date", and "end_date".
    new_booking : dict
        A dictionary representing the new booking with keys "gpu", "start_date", and "end_date".
    gpu_specs : list of dict
        A list of GPU specifications where each specification is represented as a dictionary
        with keys "name", "replicas", and "count".

    Returns
    -------
    overlapping_time_points : list of datetime
        A list of time points where bookings overlap.
    gpu_availability_per_slot : list of int
        A list of available GPU counts for each overlapping time slot.
    total_gpus : int
        The total number of GPUs available for the specified GPU type.
    """

    gpu_name = new_booking["gpu"]

    gpu_count = 0
    gpu_replicas = 0

    for gpu_spec in gpu_specs:
        if gpu_spec["name"] == gpu_name:

            gpu_replicas = gpu_spec["replicas"]
            gpu_count = gpu_spec["count"]

    overlapping_allocations = []
    for existing_booking in global_existing_bookings:
        if existing_booking["gpu"] == gpu_name:
            if isinstance(existing_booking["start_date"], str):
                existing_booking_start = datetime.strptime(existing_booking["start_date"], "%Y-%m-%d %H:%M:%S")
            else:
                existing_booking_start = existing_booking["start_date"]

            if isinstance(existing_booking["end_date"], str):
                existing_booking_end = datetime.strptime(existing_booking["end_date"], "%Y-%m-%d %H:%M:%S")
            else:
                existing_booking_end = existing_booking["end_date"]

            new_booking_start = datetime.strptime(new_booking["starting_time"], "%Y-%m-%d %H:%M:%S").replace(
                tzinfo=existing_booking_start.tzinfo
            )
            new_booking_end = datetime.strptime(new_booking["ending_time"], "%Y-%m-%d %H:%M:%S").replace(
                tzinfo=existing_booking_end.tzinfo
            )

            if new_booking_start >= existing_booking_end or new_booking_end <= existing_booking_start:
                continue
            if new_booking_start <= existing_booking_start and new_booking_end >= existing_booking_start:
                overlapping_allocations.append([existing_booking_start, existing_booking_end])
            elif new_booking_start <= existing_booking_start and new_booking_end <= existing_booking_end:
                overlapping_allocations.append([existing_booking_start, new_booking_end])
            elif new_booking_start >= existing_booking_start and new_booking_end >= existing_booking_end:
                overlapping_allocations.append([new_booking_start, existing_booking_end])
            elif new_booking_start >= existing_booking_start and new_booking_end <= existing_booking_end:
                overlapping_allocations.append([new_booking_start, new_booking_end])

    overlapping_time_points = []
    gpu_availability_per_slot = []
    for overlapping_allocation in overlapping_allocations:
        overlapping_time_points.append(overlapping_allocation[0])
        overlapping_time_points.append(overlapping_allocation[1])

    if len(global_existing_bookings) == 0:
        overlapping_time_points.append(datetime.strptime(new_booking["starting_time"], "%Y-%m-%d %H:%M:%S"))
        overlapping_time_points.append(datetime.strptime(new_booking["ending_time"], "%Y-%m-%d %H:%M:%S"))
    else:
        overlapping_time_points.append(
            datetime.strptime(new_booking["starting_time"], "%Y-%m-%d %H:%M:%S").replace(tzinfo=new_booking_start.tzinfo)
        )
        overlapping_time_points.append(
            datetime.strptime(new_booking["ending_time"], "%Y-%m-%d %H:%M:%S").replace(tzinfo=new_booking_end.tzinfo)
        )
    overlapping_time_points = sorted(set(overlapping_time_points))

    for overlapping_time_point in overlapping_time_points[:-1]:
        overlapping_window = [
            overlapping_time_point,
            overlapping_time_points[overlapping_time_points.index(overlapping_time_point) + 1],
        ]
        overlapping_window_start = overlapping_window[0]
        overlapping_window_end = overlapping_window[1]

        available_gpus = gpu_replicas * gpu_count
        gpu_availability_per_slot.append(available_gpus)
        for existing_booking in global_existing_bookings:
            if isinstance(existing_booking["start_date"], str):
                existing_booking_start = datetime.strptime(existing_booking["start_date"], "%Y-%m-%d %H:%M:%S").replace(
                    tzinfo=new_booking_start.tzinfo
                )
            else:
                existing_booking_start = existing_booking["start_date"]

            if isinstance(existing_booking["end_date"], str):
                existing_booking_end = datetime.strptime(existing_booking["end_date"], "%Y-%m-%d %H:%M:%S").replace(
                    tzinfo=new_booking_end.tzinfo
                )
            else:
                existing_booking_end = existing_booking["end_date"]

            if existing_booking_start < overlapping_window_end and existing_booking_end > overlapping_window_start:
                available_gpus -= 1
        gpu_availability_per_slot[-1] = available_gpus

    return overlapping_time_points, gpu_availability_per_slot, gpu_replicas * gpu_count


def verify_gpu_booking_policy(existing_bookings, new_booking, global_existing_bookings, gpu_specs):
    """
    Verify GPU booking policy to ensure the new booking does not exceed the allowed days and GPU availability.

    Parameters
    ----------
    existing_bookings : list
        A list of existing booking objects with `start_date` and `end_date` attributes.
    new_booking : dict
        A dictionary containing the `starting_time` and `ending_time` of the new booking in "%Y-%m-%d %H:%M:%S" format.
    global_existing_bookings : list
        A list of all existing bookings globally.
    gpu_specs : dict
        A dictionary containing the specifications of the GPUs.

    Returns
    -------
    bool
        True if the booking policy is verified, False otherwise.
    str or None
        An error message if the booking policy is not verified, None otherwise.
    """

    total_days = sum((booking.end_date - booking.start_date).days for booking in existing_bookings)

    # Calculate the number of days for the new booking
    ending_time = datetime.strptime(new_booking["ending_time"], "%Y-%m-%d %H:%M:%S")
    starting_time = datetime.strptime(new_booking["starting_time"], "%Y-%m-%d %H:%M:%S")

    new_booking_days = (ending_time - starting_time).days

    # Verify that the sum of existing bookings and the new booking does not exceed 60 days
    if total_days + new_booking_days > 60:
        return False, "The total number of days for all bookings cannot exceed 60 days."

    overlapping_time_slots, gpu_availability_per_slot, total_replicas = verify_gpu_availability(
        global_existing_bookings=global_existing_bookings, new_booking=new_booking, gpu_specs=gpu_specs
    )

    for idx, gpu_availability in enumerate(gpu_availability_per_slot):
        if gpu_availability == 0:
            error_msg = "GPU not available between the selected time slots: {} - {}".format(
                overlapping_time_slots[idx], overlapping_time_slots[idx + 1]
            )
            return False, error_msg

    return True, None


def send_maia_info_email(receiver_email, register_project_url, register_user_url, discord_support_link):
    """
    Send an email with registration information for the MAIA platform.
    Parameters
    ----------
    receiver_email : str
        The email address of the recipient.
    register_project_url : str
        The URL for project registration.
    register_user_url : str
        The URL for user registration.
    discord_support_link : str
        The URL for the MAIA support Discord.
    Returns
    -------
    None
    """

    sender_email = os.environ["email_account"]
    message = MIMEMultipart()
    message["Subject"] = "Registration Information for the MAIA Platform"
    message["From"] = "MAIA Team"
    message["To"] = receiver_email

    html = """\
    <html>
        <head></head>
        <body>
            <p>Thank you for your interest in the MAIA platform. Below are the steps to register:</p>
            <p><b>Project Registration:</b><br>
            If you are starting a research work and you want to have it hosted in MAIA, please first register your project here:<br>
            <a href="{}">MAIA Project Registration</a></p>
            <p><b>User Registration:</b><br>
            To create a user account, an active project must be available to select. Once a project is registered, you can sign up for an account linked to that project here:<br>
            <a href="{}">MAIA User Registration</a></p>
            <p>If you have any questions or need further assistance, feel free to join our Discord community:<br>
            <a href="{}">MAIA Support Discord</a></p>
            <br>
            <p>Best regards,</p>
            <p>The MAIA Admin Team</p>
        </body>
    </html>
    """.format(  # noqa: B950
        register_project_url, register_user_url, discord_support_link
    )

    # Turn these into plain/html MIMEText objects
    part1 = MIMEText(html, "html")

    message.attach(part1)

    port = 465  # For SSL
    password = os.environ["email_password"]

    # Create a secure SSL context
    context = ssl.create_default_context()

    with smtplib.SMTP_SSL(os.environ["email_smtp_server"], port, context=context) as server:
        server.login(sender_email, password)
        server.sendmail(sender_email, receiver_email, message.as_string())


def verify_minio_availability(settings):
    """
    Verifies the availability of a MinIO server.

    Parameters
    ----------
    settings : object
        An object containing the MinIO configuration settings.
        - settings.MINIO_URL : str
            The URL of the MinIO server.
        - settings.MINIO_ACCESS_KEY : str
            The access key for the MinIO server.
        - settings.MINIO_SECRET_KEY : str
            The secret key for the MinIO server.
        - settings.BUCKET_NAME : str
            The name of the bucket to check for existence.

    Returns
    -------
    bool
        True if the MinIO server is available and the bucket exists, False otherwise.
    """
    try:
        client = Minio(
            settings.MINIO_URL,
            access_key=settings.MINIO_ACCESS_KEY,
            secret_key=settings.MINIO_SECRET_KEY,
            secure=settings.MINIO_SECURE,
        )
        client.bucket_exists(settings.BUCKET_NAME)
        minio_available = True
    except Exception as e:
        print(e)
        minio_available = False

    return minio_available


def send_approved_registration_email(receiver_email, login_url, temp_password):
    """
    Sends an email to notify the user that their MAIA account registration has been approved.

    Parameters
    ----------
    receiver_email : str
        The email address of the recipient.
    login_url : str
        The URL where the user can log in to MAIA.
    temp_password : str
        The temporary password assigned to the user.

    Raises
    ------
    KeyError
        If the environment variables 'email_account' or 'email_password' are not set.
    smtplib.SMTPException
        If there is an error sending the email.
    """

    sender_email = os.environ["email_account"]
    message = MIMEMultipart()
    message["Subject"] = "Account Registration Approved"
    message["From"] = "MAIA Registration"
    message["To"] = receiver_email

    html = """\
    <html>
        <head></head>
        <body>
            <p>Your MAIA Account has been approved.</p>
            <p>Log in to MAIA at the following link: <a href="{}">MAIA</a></p>
            <p>Your temporary password is: {}</p>
            <p>Please change your password after logging in.</p>
            <br>
            <p>Best Regards,</p>
            <p>MAIA Admin Team</p>
        </body>
    </html>
    """.format(
        login_url, temp_password
    )

    # Turn these into plain/html MIMEText objects
    part1 = MIMEText(html, "html")

    message.attach(part1)

    port = 465  # For SSL
    password = os.environ["email_password"]

    # Create a secure SSL context
    context = ssl.create_default_context()

    with smtplib.SMTP_SSL(os.environ["email_smtp_server"], port, context=context) as server:
        server.login(sender_email, password)
        server.sendmail(sender_email, receiver_email, message.as_string())


def send_discord_message(username, namespace, url, project_registration=False):
    """
    Sends a message to a Discord webhook to request a MAIA account.

    Parameters
    ----------
    username : str
        The username of the person requesting the account.
    namespace : str
        The project namespace for which the account is being requested.
    url : str
        The Discord webhook URL to which the message will be sent.
    project_registration : bool, optional
        If True, indicates that a project registration is also being requested (default is False).

    Raises
    ------
    requests.exceptions.HTTPError
        If the HTTP request returned an unsuccessful status code.

    Prints
    ------
    str
        Success message with the HTTP status code if the payload is delivered successfully.
    str
        Error message if the HTTP request fails.
    """
    data = {"content": f"{username} is requesting a MAIA account for the project {namespace}.", "username": "MAIA-Bot"}

    data["embeds"] = [{"description": "MAIA User Registration Request", "title": "MAIA Account Request"}]
    if project_registration:
        data["embeds"][0]["description"] = "MAIA Project Registration Request"
        data["embeds"][0]["title"] = "MAIA Project Registration Request"
        data["content"] = f"{username} is requesting a MAIA account and a new project registration for {namespace}."

    result = requests.post(url, json=data)

    try:
        result.raise_for_status()
    except requests.exceptions.HTTPError as err:
        print(err)
    else:
        print("Payload delivered successfully, code {}.".format(result.status_code))


def get_pending_projects(settings, maia_project_model):
    """
    Retrieve a list of pending projects that are not in the active groups.

    Parameters
    ----------
    settings : dict
        Configuration settings required to access Keycloak.
    maia_project_model : Django model
        The Django model representing the MAIA projects.

    Returns
    -------
    list
        A list of namespaces of pending projects.
    """
    pending_projects = []
    try:
        active_groups = get_groups_in_keycloak(settings)

        for project in maia_project_model.objects.all():
            if project.namespace not in active_groups.values():
                pending_projects.append(project.namespace)
    except Exception:
        ...

    return pending_projects


def get_user_table(settings, maia_user_model, maia_project_model):
    """
    Retrieve user and project information from Keycloak and Minio, and organize it into a dictionary.

    Parameters
    ----------
    settings : object
        An object containing configuration settings such as OIDC and Minio credentials.
    maia_user_model : Django model
        The Django model representing MAIA users.
    maia_project_model : Django model
        The Django model representing MAIA projects.

    Returns
    -------
    tuple
        A tuple containing:
        - users_to_register_in_group (dict): Users to be registered in Keycloak groups.
        - users_to_register_in_keycloak (list): Users to be registered in Keycloak.
        - maia_group_dict (dict): Dictionary containing group information including users, conda environments, and project details.
        - users_to_remove_from_group (dict): Users to be removed from Keycloak groups.
    """

    keycloak_connection = KeycloakOpenIDConnection(
        server_url=settings.OIDC_SERVER_URL,
        username=settings.OIDC_USERNAME,
        password="",
        realm_name=settings.OIDC_REALM_NAME,
        client_id=settings.OIDC_RP_CLIENT_ID,
        client_secret_key=settings.OIDC_RP_CLIENT_SECRET,
        verify=False,
    )

    keycloak_admin = KeycloakAdmin(connection=keycloak_connection)
    users_to_register_in_group = {}

    groups = keycloak_admin.get_groups()

    maia_groups = {group["id"]: group["name"][len("MAIA:") :] for group in groups if group["name"].startswith("MAIA:")}

    pending_projects = get_pending_projects(settings=settings, maia_project_model=maia_project_model)

    maia_group_dict = {}

    try:
        client = Minio(
            settings.MINIO_URL,
            access_key=settings.MINIO_ACCESS_KEY,
            secret_key=settings.MINIO_SECRET_KEY,
            secure=settings.MINIO_SECURE,
        )

        minio_envs = [env.object_name[: -len("_env")] for env in list(client.list_objects(settings.BUCKET_NAME))]
    except Exception:
        minio_envs = []

    for maia_group in maia_groups:
        if maia_groups[maia_group] == "users":
            continue
        users = keycloak_admin.get_group_members(group_id=maia_group)

        admin_users = []
        cpu_limit = None
        memory_limit = None
        date = None
        cluster = None
        gpu = None
        environment = None
        conda_envs = []

        if maia_project_model.objects.filter(namespace=maia_groups[maia_group]).exists():
            project = maia_project_model.objects.filter(namespace=maia_groups[maia_group]).first()

            admin_users = [project.email]
            cpu_limit = project.cpu_limit
            memory_limit = project.memory_limit
            date = project.date
            cluster = project.cluster
            gpu = project.gpu
            environment = project.minimal_env

        if maia_groups[maia_group] in minio_envs:
            conda_envs.append(maia_groups[maia_group])
        else:
            conda_envs.append("N/A")

        group_users = []
        for user in users:
            if user["username"] in admin_users:
                group_users.append(user["email"] + " [Project Admin]")
            else:
                group_users.append(user["email"])

        maia_group_dict[maia_groups[maia_group]] = {
            "users": group_users,
            "conda": conda_envs,
            "admin_users": admin_users,
            "cpu_limit": cpu_limit,
            "memory_limit": memory_limit,
            "date": date,
            "cluster": cluster,
            "gpu": gpu,
            "environment": environment,
        }

    for pending_project in pending_projects:
        conda_envs = []
        if pending_project in minio_envs:
            conda_envs.append(pending_project)
        else:
            conda_envs.append("N/A")

        users = []

        for user in maia_user_model.objects.all():
            if user.namespace and pending_project in user.namespace.split(","):
                users.append(user.email)

        project = maia_project_model.objects.filter(namespace=pending_project).first()
        maia_group_dict[pending_project] = {
            "users": users,
            "pending": True,
            "conda": conda_envs,
            "admin_users": [],
            "cpu_limit": project.cpu_limit,
            "memory_limit": project.memory_limit,
            "date": project.date,
            "cluster": "N/A",
            "gpu": project.gpu,
            "environment": "Minimal",
        }

    users_to_register_in_keycloak = []
    users_to_remove_from_group = {}

    for user in maia_user_model.objects.all():
        user_groups = []
        user_in_keycloak = False
        for keycloak_user in keycloak_admin.get_users():

            if "email" in keycloak_user:
                if keycloak_user["email"] == user.email:
                    user_in_keycloak = True
                    user_keycloak_groups = keycloak_admin.get_user_groups(user_id=keycloak_user["id"])
                    for user_keycloak_group in user_keycloak_groups:
                        if user_keycloak_group["name"].startswith("MAIA:"):
                            user_groups.append(user_keycloak_group["name"][len("MAIA:") :])
        if not user_in_keycloak:
            users_to_register_in_keycloak.append(user.email)

        requested_namespaces = user.namespace.split(",") if user.namespace else []
        for requested_namespace in requested_namespaces:
            if requested_namespace not in user_groups and requested_namespace != "N/A":
                if user.email not in users_to_register_in_group:
                    users_to_register_in_group[user.email] = [requested_namespace]
                else:
                    users_to_register_in_group[user.email].append(requested_namespace)
            for user_group in user_groups:
                if user_group not in requested_namespaces:
                    if user.email not in users_to_remove_from_group:
                        users_to_remove_from_group[user.email] = [user_group]
                    else:
                        users_to_remove_from_group[user.email].append(user_group)

    return users_to_register_in_group, users_to_register_in_keycloak, maia_group_dict, users_to_remove_from_group


def register_cluster_for_project_in_db(project_model, settings, namespace, cluster):
    """
    Registers a cluster for a project in the database.
    This function connects to Keycloak to retrieve group information and
    associates a cluster with a project based on the provided namespace.
    If the project already exists, it updates the cluster information;
    otherwise, it creates a new project entry with the specified cluster.

    Parameters
    ----------
    project_model : Django model
        The Django model representing the project.
    settings : object
        An object containing configuration settings.
    namespace : str
        The namespace associated with the project.
    cluster : str
        The cluster to be registered for the project.

    Returns
    -------
    None
    """

    keycloak_connection = KeycloakOpenIDConnection(
        server_url=settings.OIDC_SERVER_URL,
        username=settings.OIDC_USERNAME,
        password="",
        realm_name=settings.OIDC_REALM_NAME,
        client_id=settings.OIDC_RP_CLIENT_ID,
        client_secret_key=settings.OIDC_RP_CLIENT_SECRET,
        verify=False,
    )

    group_id = namespace
    keycloak_admin = KeycloakAdmin(connection=keycloak_connection)
    groups = keycloak_admin.get_groups()

    maia_groups = {group["id"]: group["name"][len("MAIA:") :] for group in groups if group["name"].startswith("MAIA:")}
    for maia_group in maia_groups:
        if maia_groups[maia_group].lower().replace("_", "-") == namespace:
            group_id = maia_groups[maia_group]

    print("Registering Existing Cluster for Group: ", group_id)

    if project_model.objects.filter(namespace=group_id).exists():
        project = project_model.objects.filter(namespace=group_id).first()
        if project:
            project.cluster = cluster
            project.save()
    else:
        project_model.objects.create(namespace=group_id, cluster=cluster, memory_limit="2 Gi", cpu_limit="2")


def update_user_table(form, user_model, maia_user_model, project_model):
    """
    Updates user and project information based on the cleaned data from a form.

    Parameters
    ----------
    form : Form
        The form containing cleaned data to update the user and project models.
    user_model : Model
        The user model to query and update user information.
    maia_user_model : Model
        The MAIA user model to query and update namespace information.
    project_model : Model
        The project model to query and update project information.
    Notes
    -----
    - The function processes entries in the form's cleaned data to update user namespaces and project details.
    - User namespaces are updated or created in the `maia_user_model` based on the user's email.
    - Project details are updated or created in the `project_model` based on the namespace.
    """

    project_entries = ["memory_limit", "cpu_limit", "date", "cluster", "gpu", "minimal_environment"]

    namespace_list = []

    for entry in form.cleaned_data:
        if entry.startswith("namespace_"):
            user = user_model.objects.filter(email=entry.replace("namespace_", "")).first()
            if user:
                user_id = user.id
                if maia_user_model.objects.filter(id=user_id).exists():
                    namespace = form.cleaned_data[entry]
                    # namespaces = []
                    # for namespace in namespace_list:
                    #    if namespace.endswith(" (Pending)"):
                    #        namespaces.append(namespace[:-len(" (Pending)")])
                    #    else:
                    #        namespaces.append(namespace)
                    maia_user_model.objects.filter(id=user_id).update(namespace=namespace)
                else:
                    if user_id is not None:
                        if user_model.objects.filter(id=user_id).exists():
                            namespace = form.cleaned_data[entry]
                            # namespaces = []
                            # for namespace in namespace_list:
                            #    if namespace.endswith(" (Pending)"):
                            #        namespaces.append(namespace[:-len(" (Pending)")])
                            #    else:
                            #        namespaces.append(namespace)
                            maia_user_model.objects.create(id=user_id, namespace=namespace)
                        else:
                            namespace = form.cleaned_data[entry]
                            # namespaces = []
                            # for namespace in namespace_list:
                            #    if namespace.endswith(" (Pending)"):
                            #        namespaces.append(namespace[:-len(" (Pending)")])
                            #    else:
                            #        namespaces.append(namespace)
                            maia_user_model.objects.create(id=user_id, namespace=namespace)
        for project_entry in project_entries:
            if entry.startswith(project_entry + "_"):
                namespace = entry[len(project_entry + "_") :]
                namespace_list.append(namespace)

    for namespace in namespace_list:
        # namespaced_entries = [entry for entry in form.cleaned_data if entry.endswith(namespace)]
        if project_model.objects.filter(namespace=namespace).exists():
            project_model.objects.filter(namespace=namespace).update(
                memory_limit=form.cleaned_data["memory_limit_" + namespace],
                cpu_limit=form.cleaned_data["cpu_limit_" + namespace],
                date=form.cleaned_data["date_" + namespace],
                cluster=form.cleaned_data["cluster_" + namespace],
                gpu=form.cleaned_data["gpu_" + namespace],
                minimal_env=form.cleaned_data["minimal_environment_" + namespace],
            )
        else:
            project_model.objects.create(
                namespace=namespace,
                memory_limit=form.cleaned_data["memory_limit_" + namespace],
                cpu_limit=form.cleaned_data["cpu_limit_" + namespace],
                date=form.cleaned_data["date_" + namespace],
                cluster=form.cleaned_data["cluster_" + namespace],
                gpu=form.cleaned_data["gpu_" + namespace],
                minimal_env=form.cleaned_data["minimal_environment_" + namespace],
            )


def get_project(group_id, settings, maia_project_model, is_namespace_style=False):
    """
    Retrieve project details and associated cluster ID based on the group ID.

    Parameters
    ----------
    group_id : str
        The ID of the group to search for.
    settings : object
        The settings object containing OIDC configuration.
    maia_project_model : Django model
        The Django model representing MAIA projects.
    is_namespace_style : bool, optional
        Flag indicating whether the group ID is in namespace style (default is False).

    Returns
    -------
    tuple
        A tuple containing:
        - namespace_form (dict): A dictionary with project details and resource limits.
        - cluster_id (str or None): The ID of the associated cluster, or None if not applicable.
    """

    cluster_id = None

    for project in maia_project_model.objects.all():
        if is_namespace_style:
            if str(project.namespace).lower().replace("_", "-") == group_id:
                keycloak_connection = KeycloakOpenIDConnection(
                    server_url=settings.OIDC_SERVER_URL,
                    username=settings.OIDC_USERNAME,
                    password="",
                    realm_name=settings.OIDC_REALM_NAME,
                    client_id=settings.OIDC_RP_CLIENT_ID,
                    client_secret_key=settings.OIDC_RP_CLIENT_SECRET,
                    verify=False,
                )

                keycloak_admin = KeycloakAdmin(connection=keycloak_connection)
                groups = keycloak_admin.get_groups()

                maia_groups = {
                    group["id"]: group["name"][len("MAIA:") :] for group in groups if group["name"].startswith("MAIA:")
                }

                group_users = []

                for maia_group in maia_groups:
                    if maia_groups[maia_group] == group_id:
                        users = keycloak_admin.get_group_members(group_id=maia_group)

                        for user in users:
                            group_users.append(user["email"])

                namespace_form = {
                    "group_ID": group_id,
                    "group_subdomain": group_id.lower().replace("_", "-"),
                    "users": group_users,
                    "resources_limits": {
                        "memory": [str(int(int(project.memory_limit[: -len(" Gi")]) / 2)) + " Gi", project.memory_limit],
                        "cpu": [str(int(int(project.cpu_limit) / 2)), project.cpu_limit],
                    },
                    "environment": project.minimal_env,
                }
                if project.gpu != "N/A" and project.gpu != "NO":
                    namespace_form["gpu"] = "1"

                if project.conda != "N/A" and project.conda is not None:
                    namespace_form["minio_env_name"] = group_id + "_env"

                cluster_id = project.cluster
                if cluster_id == "N/A":
                    cluster_id = None
                return namespace_form, cluster_id
        else:
            if project.namespace == group_id:
                keycloak_connection = KeycloakOpenIDConnection(
                    server_url=settings.OIDC_SERVER_URL,
                    username=settings.OIDC_USERNAME,
                    password="",
                    realm_name=settings.OIDC_REALM_NAME,
                    client_id=settings.OIDC_RP_CLIENT_ID,
                    client_secret_key=settings.OIDC_RP_CLIENT_SECRET,
                    verify=False,
                )

                keycloak_admin = KeycloakAdmin(connection=keycloak_connection)
                groups = keycloak_admin.get_groups()

                maia_groups = {
                    group["id"]: group["name"][len("MAIA:") :] for group in groups if group["name"].startswith("MAIA:")
                }

                group_users = []

                for maia_group in maia_groups:
                    if maia_groups[maia_group] == group_id:
                        users = keycloak_admin.get_group_members(group_id=maia_group)

                        for user in users:
                            group_users.append(user["email"])

                namespace_form = {
                    "group_ID": group_id,
                    "group_subdomain": group_id.lower().replace("_", "-"),
                    "users": group_users,
                    "resources_limits": {
                        "memory": [str(int(int(project.memory_limit[: -len(" Gi")]) / 2)) + " Gi", project.memory_limit],
                        "cpu": [str(int(int(project.cpu_limit) / 2)), project.cpu_limit],
                    },
                    "environment": project.minimal_env,
                }
                if project.gpu != "N/A" and project.gpu != "NO":
                    namespace_form["gpu_request"] = "1"

                if project.conda != "N/A" and project.conda is not None:
                    namespace_form["minio_env_name"] = group_id + "_env"

                cluster_id = project.cluster
                if cluster_id == "N/A":
                    cluster_id = None
                return namespace_form, cluster_id

    return None, None


def get_argocd_project_status(argocd_namespace, project_id):
    return verify_installed_maia_toolkit(project_id=project_id, namespace=argocd_namespace, get_chart_metadata=False)


def get_allocation_date_for_project(maia_project_model, group_id, is_namespace_style=False):
    """
    Retrieves the allocation date for a project based on the given group ID.

    Parameters
    ----------
    maia_project_model : Model
        The Django model representing the MAIA project.
    group_id : str
        The group ID to match against the project's namespace.
    is_namespace_style : bool, optional
        If True, the group ID comparison will be done in a namespace style
        (lowercase and underscores replaced with hyphens). Default is False.

    Returns
    -------
    date or None
        The allocation date of the project if a match is found, otherwise None.
    """

    for project in maia_project_model.objects.all():
        if is_namespace_style:
            if str(project.namespace).lower().replace("_", "-") == group_id:
                return project.date
        else:
            if project.namespace == group_id:
                return project.date

    return None


async def get_list_of_deployed_projects():
    """
    Asynchronously retrieves a list of deployed projects from the Argo CD namespace.
    This function uses a Kubernetes client to list all releases in the "argocd" namespace
    and returns the names of these releases.

    Returns
    -------
    list of str
        A list containing the names of the deployed projects.

    Raises
    ------
    KeyError
        If the "KUBECONFIG" environment variable is not set.
    kubernetes.client.exceptions.ApiException
        If there is an error communicating with the Kubernetes API.
    """

    client = Client(kubeconfig=os.environ["KUBECONFIG"])

    releases = await client.list_releases(namespace="argocd")

    return [release.name for release in releases]


def get_project_argo_status_and_user_table(request, settings, maia_user_model, maia_project_model):
    """
    Retrieves the Argo CD project status and user table information.

    Parameters
    ----------
    request : HttpRequest
        The HTTP request object containing session and user information.
    settings : Settings
        The settings object containing configuration values.
    maia_user_model : Model
        The Django model representing MAIA users.
    maia_project_model : Model
        The Django model representing MAIA projects.

    Returns
    -------
    tuple
        A tuple containing:
        - user_table (dict): The user table information.
        - to_register_in_groups (list): List of users to register in groups.
        - to_register_in_keycloak (list): List of users to register in Keycloak.
        - maia_groups_dict (dict): Dictionary of MAIA groups.
        - project_argo_status (dict): Dictionary containing the Argo CD project status for each project.
    """
    argocd_cluster_id = settings.ARGOCD_CLUSTER

    id_token = request.session.get("oidc_id_token")
    kubeconfig_dict = generate_kubeconfig(id_token, request.user.username, "default", argocd_cluster_id, settings=settings)
    config.load_kube_config_from_dict(kubeconfig_dict)
    with open(Path("/tmp").joinpath("kubeconfig-argo"), "w") as f:
        yaml.dump(kubeconfig_dict, f)
        os.environ["KUBECONFIG"] = str(Path("/tmp").joinpath("kubeconfig-argo"))

    to_register_in_groups, to_register_in_keycloak, maia_groups_dict, users_to_remove_from_group = get_user_table(
        settings=settings, maia_user_model=maia_user_model, maia_project_model=maia_project_model
    )
    project_argo_status = {}

    namespaces = get_namespaces(id_token, api_urls=settings.API_URL, private_clusters=settings.PRIVATE_CLUSTERS)

    deployed_projects = asyncio.run(get_list_of_deployed_projects())
    for project_id in maia_groups_dict:
        if project_id.lower().replace("_", "-") in deployed_projects:
            project_argo_status[project_id] = 1
        else:
            project_argo_status[project_id] = -1
        if "ARGOCD_DISABLED" in os.environ and os.environ["ARGOCD_DISABLED"] == "True":
            if project_id.lower().replace("_", "-") in namespaces:
                project_argo_status[project_id] = 1
            else:
                project_argo_status[project_id] = -1
        # project_argo_status[project_id] = asyncio.run(get_argocd_project_status(argocd_namespace="argocd", project_id=project_id.lower().replace("_", "-")))  # noqa: B950

    return to_register_in_groups, to_register_in_keycloak, maia_groups_dict, project_argo_status, users_to_remove_from_group


def send_maia_message_email(receiver_emails, subject, message_body):
    """
    Send an email with a custom message to multiple recipients with improved deliverability.
    """
    try:
        sender_email = os.environ["email_account"]
        message = MIMEMultipart("alternative")  # Changed to alternative for better compatibility

        # Add proper headers to improve deliverability
        message["Subject"] = subject
        message["From"] = f"MAIA Team <{sender_email}>"  # Use proper From format
        message["To"] = ", ".join(receiver_emails)
        message["Reply-To"] = sender_email
        message["Date"] = email.utils.formatdate(localtime=True)
        message["Message-ID"] = email.utils.make_msgid(domain=sender_email.split("@")[1])

        html = f"""\
        <html>
            <head>
                <meta charset="UTF-8">
                <meta name="viewport" content="width=device-width, initial-scale=1.0">
            </head>
            <body style="font-family: Arial, sans-serif; line-height: 1.6; color: #333333;">
                {message_body}
                <br>
                <p>Best regards,</p>
                <p>The MAIA Admin Team</p>
                <hr>
                <p style="font-size: 12px; color: #666666;">
                    This is an automated message from the MAIA Platform.
                    If you believe you received this in error, please contact support.
                </p>
            </body>
        </html>
        """

        # Create plain text version
        text = BeautifulSoup(html, "html.parser").get_text()

        # Attach both plain text and HTML versions
        part1 = MIMEText(text, "plain")
        part2 = MIMEText(html, "html")
        message.attach(part1)
        message.attach(part2)

        port = 465  # For SSL
        password = os.environ["email_password"]
        smtp_server = os.environ["email_smtp_server"]

        context = ssl.create_default_context()

        with smtplib.SMTP_SSL(smtp_server, port, context=context) as server:
            server.login(sender_email, password)
            server.send_message(message)  # Using send_message instead of sendmail

        return True

    except Exception as e:
        print(f"Error sending email: {str(e)}")
        return False


def generate_encryption_keys(folder_path):
    """
    Generate RSA encryption keys and save them to files.

    Parameters
    ----------
    folder_path : str
        The path to the folder where the keys will be saved.

    Returns
    -------
    None
    """

    # Generate RSA key pair
    private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)

    # Extract public key
    public_key = private_key.public_key()

    # Save private key to a file
    with open(Path(folder_path).joinpath("private_key.pem"), "wb") as f:
        f.write(
            private_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.PKCS8,
                encryption_algorithm=serialization.NoEncryption(),
            )
        )

    # Save public key to a file
    with open(Path(folder_path).joinpath("public_key.pem"), "wb") as f:
        f.write(
            public_key.public_bytes(encoding=serialization.Encoding.PEM, format=serialization.PublicFormat.SubjectPublicKeyInfo)
        )

    print("Keys generated successfully!")


def encrypt_string(public_key, string):
    """
    Encrypts a given string using the provided public key.

    Parameters
    ----------
    public_key : str
        The file path to the public key in PEM format.
    string : str
        The string to be encrypted.

    Returns
    -------
    str
        The encrypted string in hexadecimal format.

    Raises
    ------
    ValueError
        If the public key file cannot be read or is invalid.
    """

    # Load public key
    with open(public_key, "rb") as f:
        public_key = serialization.load_pem_public_key(f.read())

    def encrypt_message(message, public_key):
        encrypted = public_key.encrypt(
            message.encode(), padding.OAEP(mgf=padding.MGF1(algorithm=hashes.SHA256()), algorithm=hashes.SHA256(), label=None)
        )
        return encrypted

    encrypted_message = encrypt_message(string, public_key)

    return encrypted_message.hex()


def decrypt_string(private_key, string):
    """
    Decrypts an encrypted string using a given private key.

    Parameters
    ----------
    private_key : str
        Path to the private key file in PEM format.
    string : bytes
        The encrypted string to be decrypted.

    Returns
    -------
    str
        The decrypted string.

    Raises
    ------
    ValueError
        If the decryption process fails.
    """

    with open(private_key, "rb") as f:
        private_key = serialization.load_pem_private_key(f.read(), password=None)

    def decrypt_message(encrypted, private_key):
        decrypted = private_key.decrypt(
            encrypted, padding.OAEP(mgf=padding.MGF1(algorithm=hashes.SHA256()), algorithm=hashes.SHA256(), label=None)
        )
        return decrypted.decode()

    decrypted_message = decrypt_message(string, private_key)

    return decrypted_message



#!/usr/bin/env python

from __future__ import annotations

import argparse
import datetime
import os
import smtplib
import ssl
import sys
from argparse import RawTextHelpFormatter
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from pathlib import Path
from textwrap import dedent

import dotenv

import MAIA

version = MAIA.__version__

TIMESTAMP = "{:%Y-%m-%d_%H-%M-%S}".format(datetime.datetime.now())


dotenv.load_dotenv()

EPILOG = dedent(
    """
    Example call:
    ::
        {filename} --email <recipient_email> --url <maia_platform_url>
    """.format(  # noqa: E501
        filename=Path(__file__).stem
    )
)


def get_arg_parser():
    parser = argparse.ArgumentParser(
        description="Send welcome email to new MAIA users", epilog=EPILOG, formatter_class=RawTextHelpFormatter
    )
    parser.add_argument("--email", required=True, help="Recipient email address")
    parser.add_argument("--url", required=True, help="MAIA platform URL")
    parser.add_argument("-v", "--version", action="version", version="%(prog)s " + version)

    return parser


def send_welcome_user_email(receiver_email, maia_url):
    """
    Send a welcome email to new MAIA users with platform information and getting started resources.

    Parameters
    ----------
    receiver_email : str
        The email address of the recipient
    maia_url : str
        The base URL of the MAIA platform
    """

    sender_email = os.environ["email_account"]
    message = MIMEMultipart()
    message["Subject"] = "Welcome to the MAIA Platform"
    message["From"] = "MAIA Team"
    message["To"] = receiver_email

    html = """\
    <html>
        <head></head>
        <body>
            <p>Welcome to MAIA!</p>
            <p>We're excited to have you join our platform. Your login details are your KTH credentials (unless otherwise provided), and you need to setup multifactor authentication to use the platform.
            Your MAIA account has been created and you can now access the platform at:<br>
            <a href="{}">{}</a></p>
            You can also setup the ssh connection here.

            <p>To see you workspace details and find the ssh connection and jupyterhub details, please visit the MAIA Dashboard at:<br>
            <a href="https://maia.app.cloud.cbh.kth.se/maia/">https://maia.app.cloud.cbh.kth.se/maia/</a></p>

            <p><b>Getting Started:</b></p>
            <ul>
                <li>Sign in to the MAIA Dashboard to access your projects</li>
                <li>Access your project workspace through JupyterLab</li>
                <li>Install additional packages using conda or pip</li>
            </ul>

            <p><b>Key Features:</b></p>
            <ul>
                <li>GPU support for machine learning workloads</li>
                <li>Persistent storage for your research data</li>
            </ul>

            <p><b>Resources:</b></p>
            <ul>
                <li>Support: Join our Discord community at, support in maia-support channel <a href="https://discord.gg/ZSe2dDzt">https://discord.gg/ZSe2dDzt</a></li>
                <li>Tutorial notebooks are available in your workspace</li>
            </ul>

            <br>
            <p>Best regards,</p>
            <p>The MAIA Admin Team</p>
        </body>
    </html>
    """.format(  # noqa: E501, B950
        maia_url, maia_url
    )

    part1 = MIMEText(html, "html")
    message.attach(part1)

    port = 465  # For SSL
    password = os.environ["email_password"]

    # Create a secure SSL context
    context = ssl.create_default_context()

    with smtplib.SMTP_SSL(os.environ["email_smtp_server"], port, context=context) as server:
        server.login(sender_email, password)
        server.sendmail(sender_email, receiver_email, message.as_string())


def main():

    args = get_arg_parser().parse_args()
    try:
        send_welcome_user_email(args.email, args.url)
        print(f"Welcome email sent successfully to {args.email}")
    except Exception as e:
        print(f"Error sending welcome email: {str(e)}", file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()




# This file was generated by 'versioneer.py' (0.29) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2025-07-29T09:11:22+0200",
 "dirty": false,
 "error": null,
 "full-revisionid": "5672b8f5f34294613aa85748ebd61d02db199de4",
 "version": "1.9a"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)



import os

import setuptools
from setuptools import setup

import versioneer


def resolve_requirements(file):
    requirements = []
    with open(file) as f:
        req = f.read().splitlines()
        for r in req:
            if r.startswith("-r"):
                requirements += resolve_requirements(os.path.join(os.path.dirname(file), r.split(" ")[1]))
            else:
                requirements.append(r)
    return requirements


def read_file(file):
    with open(file) as f:
        content = f.read()
    return content


setup(
    version=versioneer.get_version(),
    packages=setuptools.find_packages(),
    cmdclass=versioneer.get_cmdclass(),
    zip_safe=False,
    data_files=[('', ["requirements.txt"]), ],
    package_data={
        "": ["configs/*.yaml", "configs/*.json", "configs/*.yml"],
    },
    entry_points={
        "console_scripts": [
            "MAIA_deploy_helm_chart = MAIA_scripts.MAIA_deploy_helm_chart:main",
            "MAIA_install_admin_toolkit = MAIA_scripts.MAIA_install_admin_toolkit:main",
            "MAIA_install_project_toolkit = MAIA_scripts.MAIA_install_project_toolkit:main",
            "MAIA_install_core_toolkit = MAIA_scripts.MAIA_install_core_toolkit:main",
            "MAIA_create_JupyterHub_config = MAIA_scripts.MAIA_create_JupyterHub_config:main",
            "MAIA_send_welcome_user_email = MAIA_scripts.MAIA_send_welcome_user_email:main",
            "MAIA_send_all_user_email = MAIA_scripts.MAIA_send_all_user_email:main",
        ],
    },
    keywords=["helm", "kubernetes", "maia", "resource deployment"],

)
